/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief pthread_brrr_t are functions that provide a waiting mechanism.
 *
 * A barrier causes the calling thread to wait until a set number of
 * threads reach a point, then those threads may continue.
 * @warning <ul><li>not async-cancel safe</li></ul>
 * @warning <ul><li>cancellation has not been tested</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_BRRR_
#define _PTHREAD_BRRR_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief return value for last barrier thread
 */
#define PTHREAD_BRRR_SERIAL_THREAD -1

/**
 * @brief pthread_brrr_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

} pthread_brrrattr_t;

/**
 * @brief pthread_brrr_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief wait condition
     */
    pthread_cond_t cond;

    /**
     * @brief  current cycle barrier is in
     */
    int cycle;

    /**
     * @brief current count of threads waiting
     */
    int count;

    /**
     * @brief number of threads to wait for
     */
    int max_count;

#ifdef CTF_STATS
    /**
     * @brief count of threads that entered the barrier
     */
    int total_count;

    /**
     * @brief count of barrier waited for
     */
    int total_barriers;

    /**
     * @brief sum of time waiting to exit barrier (sec)
     */
    float total_wait;

#endif

} pthread_brrr_t;


/**
 * @brief static initializer for pthread_brrr_t
 */
#ifdef CTF_STATS
#define PTHREAD_BRRR_INITIALIZER(count) { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, 0, 0, count, \
        0, 0, 0.0 }
#else
#define PTHREAD_BRRR_INITIALIZER(count) { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, 0, 0, count }
#endif

/**
 * @brief initialize pthread_brrr_t structure
 * @param brrr pointer to pthread_brrr_t structure
 * @param attr pointer to pthread_brrr_t attribute structure
 * @param value number of thread barriers to wait for
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrr_init(pthread_brrr_t * brrr,
        const pthread_brrrattr_t * attr, const int value);

/**
 * @brief destroy pthread_brrr_t structure
 * @param brrr pointer to pthread_brrr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrr_destroy(pthread_brrr_t * brrr);

/**
 * @brief to wait until the previously set number of waiters reach this barrier
 * @param brrr pointer to pthread_brrr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrr_wait(pthread_brrr_t * brrr);

/**
 * @brief to decrement the number of threads to wait for
 * @param brrr pointer to pthread_brrr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrr_decrement_np(pthread_brrr_t * brrr);

/**
 * @brief to increment the number of threads to wait for
 * @param brrr pointer to pthread_brrr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrr_increment_np(pthread_brrr_t * brrr);
/*
    arguments:
        (1) pointer to an initialized pthread_brrr_t [input/output]
     functionality:
        (1) to increment the number of threads to wait for
*/

/**
 * @brief initialize pthread_brrr_t attribute structure
 * @param attr pointer to pthread_brrr_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrrattr_init(pthread_brrrattr_t * attr);

/**
 * @brief destroy pthread_brrr_t attribute structure
 * @param attr pointer to pthread_brrr_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrrattr_destroy(pthread_brrrattr_t * attr);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief set pshared to current process sharing value
 * @param attr pointer to pthread_brrr_t attribute structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrrattr_getpshared(const pthread_brrrattr_t * attr,
        int * pshared);

/**
 * @brief set current process sharing value to pshared
 * @param attr pointer to pthread_brrr_t attribute structure
 * @param pshared desired process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_brrrattr_setpshared(pthread_brrrattr_t * attr,
        const int pshared);
#endif

#ifdef __cplusplus
}
#endif

#endif
