/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief pthread_fifo_mutex_t are functions that provide a first-in first-out mutex
 *
 * @warning <ul><li>not async-cancel safe</li></ul>
 * @warning <ul><li>cancellation is disabled in pthread_fifo_mutex_lock()</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_FIFO_MUTEX_
#define _PTHREAD_FIFO_MUTEX_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief enumeration for fifo objects
 */
enum
{
    /**
     * @brief non-blocking enumeration
     */
    PTHREAD_FIFO_MUTEX_NONBLOCKING,

    /**
     * @brief blocking enumeration
     */
    PTHREAD_FIFO_MUTEX_BLOCKING,

    /**
     * @brief default blocking enumeration
     */
    PTHREAD_FIFO_MUTEX_DEFAULT = PTHREAD_FIFO_MUTEX_BLOCKING
};

/**
 * @brief non-blocking enumeration
 */
#define PTHREAD_FIFO_MUTEX_NONBLOCKING PTHREAD_FIFO_MUTEX_NONBLOCKING

/**
 * @brief blocking enumeration
 */
#define PTHREAD_FIFO_MUTEX_BLOCKING PTHREAD_FIFO_MUTEX_BLOCKING

/**
 * @brief default blocking enumeration
 */
#define PTHREAD_FIFO_MUTEX_DEFAULT PTHREAD_FIFO_MUTEX_DEFAULT


/**
 * @brief pthread_fifo_mutex_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

    /**
     * @brief blocking attribute
     */
    int block;

} pthread_fifo_mutexattr_t;


/**
 * @brief pthread_fifo_mutex_t data structure
 */
struct fifo_data_t
{
    /**
     * @brief  pthread_t ID of waiting thread
     */
    pthread_t tid;

    /**
     * @brief condition to wake up waiting thread
     */
    pthread_cond_t cond;
};


/**
 * @brief pthread_fifo_mutex_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief fifo queue not full condition
     */
    pthread_cond_t cond;

    /**
     * @brief fifo queue array
     */
    struct fifo_data_t * queue;

    /**
     * @brief index of front of fifo queue
     */
    int front;

    /**
     * @brief index of back of fifo queue
     */
    int back;

    /**
     * @brief fifo queue length
     */
    int length;

    /**
     * @brief  maximum fifo queue length
     */
    int max_length;

    /**
     * @brief count of threads blocking
     */
    int queue_full_wait;

    /**
     * @brief fifo-mutex locked flag
     */
    short locked;

    /**
     * @brief blocking attribute
     */
    int block;

#ifdef CTF_STATS
    /**
     * @brief total pass count for fifo-mutex
     */
    int total_pass_count;

    /**
     * @brief total wait count for fifo-mutex
     */
    int total_wait_count;

    /**
     * @brief sum of time waited (sec)
     */
    float total_wait;
    /* sum of time waited (sec) */

    /**
     * @brief sum of time locked (sec)
     */
    float total_locked;
#endif

} pthread_fifo_mutex_t;


/**
 * @brief initialize pthread_fifo_mutex_t structure
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @param attr pointer to pthread_fifo_mutex_t attribute structure
 * @param length maximum length of fifo
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_init(pthread_fifo_mutex_t * mutex,
        const pthread_fifo_mutexattr_t * attr, const int length);

/**
 * @brief destroy pthread_fifo_mutex_t structure
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_destroy(pthread_fifo_mutex_t * mutex);

/**
 * @brief suspend calling thread until lock is available
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_lock(pthread_fifo_mutex_t * mutex);

/**
 * @brief if mutex is not lock then obtain lock
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_trylock(pthread_fifo_mutex_t * mutex);

/**
 * @brief if mutex is not lock then obtain lock or return after abstime
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @param abstime pointer to timespec structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_timedlock(pthread_fifo_mutex_t * mutex,
        const struct timespec * abstime);

/**
 * @brief unlock mutex
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_unlock(pthread_fifo_mutex_t * mutex);

#if (_POSIX_THREAD_PRIO_PROTECT == 1)
/**
 * @brief assign prioceiling the current priority ceiling value
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @param prioceiling pointer to priority ceiling
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_getprioceiling(const pthread_fifo_mutex_t * mutex,
        int * prioceiling);

/**
 * @brief assign old_ceiling the current priority ceiling value and
 * set calling thread priority ceiling to prioceiling
 * @param mutex pointer to pthread_fifo_mutex_t structure
 * @param prioceiling priority ceiling
 * @param old_ceiling pointer to previous priority ceiling
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutex_setprioceiling(pthread_fifo_mutex_t * mutex,
        const int prioceiling, int * old_ceiling);
#endif

/**
 * @brief initialize pthread_fifo_mutex_t attribute structure
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_init(pthread_fifo_mutexattr_t * attr);

/**
 * @brief destroy pthread_fifo_mutex_t attribute structure
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_destroy(pthread_fifo_mutexattr_t * attr);

/**
 * @brief assign block the current blocking attribute
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param block pointer to blocking attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_getblock(const pthread_fifo_mutexattr_t * attr,
        int * block);

/**
 * @brief assign the current blocking attribute
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param block blocking attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_setblock(pthread_fifo_mutexattr_t * attr,
        const int block);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief assign pshared the current process sharing value
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_getpshared(const pthread_fifo_mutexattr_t * attr,
        int * pshared);

/**
 * @brief set current process sharing value to pshared
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param pshared process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_setpshared(pthread_fifo_mutexattr_t * attr,
        const int pshared);
#endif

#if (_POSIX_THREAD_PRIO_PROTECT == 1)
/**
 * @brief assign prioceiling the current priority ceiling value
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param prioceiling pointer to priority ceiling
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_getprioceiling(const pthread_fifo_mutexattr_t * attr,
        int * prioceiling);

/**
 * @brief set calling thread priority ceiling to prioceiling
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param prioceiling priority ceiling
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_setprioceiling(pthread_fifo_mutexattr_t * attr,
        const int prioceiling);

#if (_POSIX_THREAD_PRIO_INHERIT == 1)
/**
 * @brief assign protocol the current protocol value
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param protocol pointer to protocol value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_getprotocol(const pthread_fifo_mutexattr_t * attr,
        int * protocol);

/**
 * @brief set calling thread protocol value to protocol
 * @param attr pointer to pthread_fifo_mutex_t structure
 * @param protocol protocol value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_fifo_mutexattr_setprotocol(pthread_fifo_mutexattr_t * attr,
        const int protocol);
#endif
#endif

#ifdef __cplusplus
}
#endif

#endif
