/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief pthread_jail_t provides a multiple thread update to a dynamic array
 *
 * to enter the jail:
 *   -# must get searched (only one at a time)
 *   -# if case (test array length)<br>
 *      -# enter lock-down (no one else can be in jail)
 *      -# update (maximum array length)
 *      -# release from lock-down
 *   -# update (current array length)
 *   -# become visitor (update array)
 *   -# leave jail
 *
 * @warning <ul><li>not async-cancel safe</li></ul>
 * @warning <ul><li>cancellation has not been tested</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_JAIL_
#define _PTHREAD_JAIL_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief pthread_jail_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

} pthread_jailattr_t;


/**
 * @brief pthread_jail_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief search condition
     */
    pthread_cond_t search_cond;

    /**
     * @brief lock condition
     */
    pthread_cond_t lock_cond;

    /**
     * @brief number of visitors
     */
    int visit_count;

    /**
     * @brief number of searchers
     */
    int search_count;

    /**
     * @brief number of searchers waiting
     */
    int search_wait;

    /**
     * @brief number of lockers
     */
    int lock_count;

    /**
     * @brief number of lockers waiting
     */
    int lock_wait;

#ifdef CTF_STATS
    /**
     * @brief total count of visitors
     */
    int total_visitor_count;

    /**
     * @brief total count of waits
     */
    int total_wait_count;

    /**
     * @brief total count of lockdowns
     */
    int total_lockdown_count;

    /**
     * @brief sum of time searchers waited (sec)
     */
    float total_search_wait;

    /**
     * @brief sum of time waited for a lockdown (sec)
     */
    float total_lockdown_wait;

    /**
     * @brief sum of time waited for a lockdown (sec)
     */
    float total_lockdown;
#endif

} pthread_jail_t;


/**
 * @brief static initializer for pthread_jail_t
 */
#ifdef CTF_STATS
#define PTHREAD_JAIL_INITIALIZER { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER, \
        0, 0, 0, 0, 0, \
        0, 0, 0, 0.0, 0.0, 0.0 }
#else
#define PTHREAD_JAIL_INITIALIZER { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER, \
        0, 0, 0, 0, 0 }
#endif


/**
 * @brief initialize pthread_jail_t structure
 * @param jail pointer to pthread_jail_t structure
 * @param attr pointer to pthread_jail_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_init(pthread_jail_t * jail,
        const pthread_jailattr_t * attr);

/**
 * @brief destroy pthread_jail_t structure
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_destroy(pthread_jail_t * jail);

/**
 * @brief only one can be searched at a time
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_search(pthread_jail_t * jail);

/**
 * @brief multiple visitors at a time
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_visit(pthread_jail_t * jail);

/**
 * @brief leaving jail
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_leave(pthread_jail_t * jail);

/**
 * @brief no new visitors, wait until current visitors leave
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_lockdown(pthread_jail_t * jail);

/**
 * @brief new visitors allowed
 * @param jail pointer to pthread_jail_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jail_release(pthread_jail_t * jail);

/**
 * @brief initialize pthread_jail_t attribute structure
 * @param attr pointer to pthread_jail_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jailattr_init(pthread_jailattr_t * attr);

/**
 * @brief destroy pthread_jail_t attribute structure
 * @param attr pointer to pthread_jail_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jailattr_destroy(pthread_jailattr_t * attr);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief set pshared to current process sharing value
 * @param attr pointer to pthread_jail_t attribute structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jailattr_getpshared(const pthread_jailattr_t * attr,
        int * pshared);

/**
 * @brief set current process sharing value to pshared
 * @param attr pointer to pthread_jail_t attribute structure
 * @param pshared desired process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_jailattr_setpshared(pthread_jailattr_t * attr,
        const int pshared);
#endif


#ifdef __cplusplus
}
#endif

#endif
