/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief provides a repeatable threaded two dimensional for loop construct
 *
 * @warning <ul><li>must cancel all threads using mgate</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/07/17
 * @version 1.0
 */


#ifndef _PTHREAD_MGATE_
#define _PTHREAD_MGATE_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief pthread_mgate_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

} pthread_mgateattr_t;


/**
 * @brief pthread_mgate_t data structure
 */
struct mgate_data_t
{
    /**
     * @brief order index
     */
    int order;

    /**
     * @brief gate index
     */
    int gate;

    /**
     * @brief wait index
     */
    int wait;

    /**
     * @brief no wait token flag
     */
    short token;

    /**
     * @brief gate condition
     */
    pthread_cond_t cond;
};


/**
 * @brief pthread_mgate_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief array of gates
     */
    struct mgate_data_t * data;

    /**
     * @brief number of gates
     */
    int num_gates;

    /**
     * @brief next index for order
     */
    int next_order;

#ifdef CTF_STATS
    /**
     * @brief total count passed without waiting
     */
    int total_pass_count;

    /**
     * @brief total count passed after waiting
     */
    int total_wait_count;

    /**
     * @brief sum of time waited (sec)
     */
    float total_wait;
#endif

} pthread_mgate_t;


/**
 * @brief initialize pthread_mgate_t structure
 * @param mgate pointer to pthread_mgate_t structure
 * @param attr pointer to pthread_mgate_t attribute structure
 * @param num_gates number of gates
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_init(pthread_mgate_t * mgate,
        const pthread_mgateattr_t * attr, const int num_gates);

/**
 * @brief destroy pthread_mgate_t structure
 * @param mgate pointer to pthread_mgate_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_destroy(pthread_mgate_t * mgate);

/**
 * @brief wait until gate is open
 * @param mgate pointer to pthread_mgate_t structure
 * @param gate_index gate index ID
 * @param check_index check index
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_check(pthread_mgate_t * mgate, const int gate_index,
        const int check_index);

/**
 * @brief update next gate to new values
 * @param mgate pointer to pthread_mgate_t structure
 * @param gate_index gate index ID
 * @param update_index update index
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_update(pthread_mgate_t * mgate,
        const int gate_index, const int update_index);

/**
 * @brief update to next gate
 * @param mgate pointer to pthread_mgate_t structure
 * @param gate_index gate index ID
 * @param order_number pointer to order number
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_pass(pthread_mgate_t * mgate,
        const int gate_index, int * order_number);

/**
 * @brief reset gate information to starting values
 * @param mgate pointer to pthread_mgate_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgate_reset(pthread_mgate_t * mgate);

/**
 * @brief initialize pthread_mgate_t attribute structure
 * @param attr pointer to pthread_mgate_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgateattr_init(pthread_mgateattr_t * attr);

/**
 * @brief destroy pthread_mgate_t attribute structure
 * @param attr pointer to pthread_mgate_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgateattr_destroy(pthread_mgateattr_t * attr);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief set pshared to current process sharing value
 * @param attr pointer to pthread_mgate_t attribute structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgateattr_getpshared(const pthread_mgateattr_t * attr,
        int * pshared);

/**
 * @brief set current process sharing value to pshared
 * @param attr pointer to pthread_mgate_t attribute structure
 * @param pshared desired process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_mgateattr_setpshared(pthread_mgateattr_t * attr,
        const int pshared);
#endif

#ifdef __cplusplus
}
#endif

#endif
