/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief functions that provide a pool of threads to pass work to
 *
 * @warning <ul><li>cancellation is disabled in pool threads</li></ul>
 * @warning <ul><li>default thread creation attributes are defualt pool attributes</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_POOL_
#define _PTHREAD_POOL_


#include <pthread.h>
#include <sched.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief pthread_pool_t blocking attribute enumeration
 */
enum
{
    /**
     * @brief pthread_pool_t non-blocking enumeration
     */
    PTHREAD_POOL_NONBLOCKING,

    /**
     * @brief pthread_pool_t blocking enumeration
     */
    PTHREAD_POOL_BLOCKING
};

/**
 * @brief pthread_pool_t non-blocking enumeration
 */
#define PTHREAD_POOL_NONBLOCKING PTHREAD_POOL_NONBLOCKING

/**
 * @brief pthread_pool_t blocking enumeration
 */
#define PTHREAD_POOL_BLOCKING PTHREAD_POOL_BLOCKING


#ifdef CTF_DEBUG
#if (defined(_POSIX_THREAD_ATTR_STACKSIZE) && \
        defined(_POSIX_THREAD_ATTR_STACKADDR))
/**
 * @brief pthread_pool_t check stack attribute enumeration
 */
enum
{
    /**
     * @brief pthread_pool_t check stack off enumeration
     */
    PTHREAD_POOL_CHECKSTACK_OFF,

    /**
     * @brief pthread_pool_t check stack on enumeration
     */
    PTHREAD_POOL_CHECKSTACK_ON
};

/**
 * @brief pthread_pool_t check stack off enumeration
 */
#define PTHREAD_POOL_CHECKSTACK_OFF PTHREAD_POOL_CHECKSTACK_OFF

/**
 * @brief pthread_pool_t check stack on enumeration
 */
#define PTHREAD_POOL_CHECKSTACK_ON PTHREAD_POOL_CHECKSTACK_ON
#endif
#endif


/**
 * @brief pthread_pool_t data structure
 */
struct pool_data_t
{
    /**
     * @brief function pointer
     */
    void (*routine)(void *);

    /**
     * @brief function argument
     */
    void * arg;
};


/**
 * @brief pthread_pool_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief thread attributes
     */
    pthread_attr_t thread_attr;

#ifdef CTF_DEBUG
#if ((_POSIX_THREAD_ATTR_STACKSIZE == 1) && \
        (_POSIX_THREAD_ATTR_STACKADDR == 1))
    /**
     * @brief check stack attribute
     */
    short check_stack;
    /* check stack attribute */
#endif
#endif

    /**
     * @brief process sharing attribute
     */
    int pshared;

    /**
     * @brief pool blocking attribute
     */
    int block;

} pthread_poolattr_t;


/**
 * @brief pthread_pool_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief queue not empty condition
     */
    pthread_cond_t queue_not_empty;

    /**
     * @brief queue not full condition
     */
    pthread_cond_t queue_not_full;

    /**
     * @brief work finished condition
     */
    pthread_cond_t work_finished;

    /**
     * @brief pointer to an array of threads
     */
    pthread_t * threads;

    /**
     * @brief number of threads in pool
     */
    int num_threads;

    /**
     * @brief pointer to an array of work
     */
    struct pool_data_t * queue;

    /**
     * @brief index of front of work data
     */
    int front;

    /**
     * @brief index of back of work data
     */
    int back;

    /**
     * @brief queue open flag
     */
    short queue_open;

    /**
     * @brief maximum queue length
     */
    int max_queue_length;

    /**
     * @brief current queue length
     */
    int queue_length;

    /**
     * @brief number of threads currently active
     */
    int num_active;

    /**
     * @brief number of threads who are destroyed
     */
    int destroyed;

    /**
     * @brief number of threads waiting for entry in queue
     */
    int queue_full_wait;

    /**
     * @brief number of threads waiting for current work to be finished
     */
    int work_not_finished_wait;

    /**
     * @brief blocking value
     */
    int block;

#ifdef CTF_DEBUG
#if ((_POSIX_THREAD_ATTR_STACKSIZE == 1) && \
        (_POSIX_THREAD_ATTR_STACKADDR == 1))
    /**
     * @brief check stack flag
     */
    short check_stack;

    /**
     * @brief pointer to array of stack addresses
     */
    size_t ** stack_address;

    /**
     * @brief thread stack size
     */
    size_t stack_size;

    /**
     * @brief thread guard size
     */
    size_t guard_size;
#endif
#endif

} pthread_pool_t;


/**
 * @brief initialize pthread_pool_t structure
 * @param pool pointer to pthread_pool_t structure
 * @param attr pointer to pthread_pool_t attribute structure
 * @param num_threads number of threads
 * @param max_queue_length maximum queue length
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_init(pthread_pool_t * pool,
        const pthread_poolattr_t * attr,
        const int num_threads, const int max_queue_length);

/**
 * @brief destroy pthread_pool_t structure
 * @param pool pointer to pthread_pool_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_destroy(pthread_pool_t * pool);

/**
 * @brief add routine(arg) to pool work to be done
 * @param pool pointer to pthread_pool_t structure
 * @param routine pointer to a function
 * @param arg pointer to argument for routine
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_addwork(pthread_pool_t * pool,
        void (*routine)(void *), void * arg);

/**
 * @brief set index to thread pool index (starting at zero)
 * @param pool pointer to pthread_pool_t structure
 * @param pool_index pointer to an index
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_getindex(pthread_pool_t * pool,
        int * pool_index);

/**
 * @brief suspend calling thread until all pool work is done
 * @param pool pointer to pthread_pool_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_finishwork(pthread_pool_t * pool);

#ifdef CTF_DEBUG
#if ((_POSIX_THREAD_ATTR_STACKSIZE == 1) && \
        (_POSIX_THREAD_ATTR_STACKADDR == 1))
/**
 * @brief display the current maximum stack size used
 * @param pool pointer to pthread_pool_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_pool_db_display_maxstack(pthread_pool_t * pool);
#endif
#endif

/**
 * @brief initialize pthread_pool_t attribute structure
 * @param attr pointer to pthread_pool_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_init(pthread_poolattr_t * attr);

/**
 * @brief destroy pthread_pool_t attribute structure
 * @param attr pointer to pthread_pool_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_destroy(pthread_poolattr_t * attr);

/**
 * @brief set block to current blocking value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param block pointer to blocking value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getblock(const pthread_poolattr_t * attr,
        int * block);

/**
 * @brief set current blocking value to block
 * @param attr pointer to pthread_pool_t attribute structure
 * @param block blocking value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setblock(pthread_poolattr_t * attr,
        const int block);

#if ((_XOPEN_VERSION - 0) >= 500)
/**
 * @brief set guardsize to current thread guard size value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param guardsize pointer to thread guard size value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getguardsize(const pthread_poolattr_t * attr,
        size_t * guardsize);

/**
 * @brief set current thread guard size value to guardsize
 * @param attr pointer to pthread_pool_t attribute structure
 * @param guardsize thread guard size value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setguardsize(pthread_poolattr_t * attr,
        const size_t guardsize);
#endif

#if (_POSIX_THREAD_PRIORITY_SCHEDULING == 1)
/**
 * @brief set inherit to current schedule inheritance value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param inherit pointer to schedule inheritance value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getsinheritsched(const pthread_poolattr_t * attr,
        int * inherit);

/**
 * @brief set current schedule inheritance value to inherit
 * @param attr pointer to pthread_pool_t attribute structure
 * @param inherit inheritance value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setsinheritsched(pthread_poolattr_t * attr,
        const int inherit);

/**
 * @brief set param to current schedule priority value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param param pointer to schedule priority value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getschedparam(const pthread_poolattr_t * attr,
        struct sched_param * param);

/**
 * @brief set current schedule priority value to param
 * @param attr pointer to pthread_pool_t attribute structure
 * @param param schedule priority value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setschedparam(pthread_poolattr_t * attr,
        const struct sched_param * param);

/**
 * @brief set policy to current schedule policy value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param policy pointer to schedule policy value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getschedpolicy(const pthread_poolattr_t * attr,
        int * policy);

/**
 * @brief set current schedule policy value to policy
 * @param attr pointer to pthread_pool_t attribute structure
 * @param policy schedule policy value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setschedpolicy(pthread_poolattr_t * attr,
        const int policy);

/**
 * @brief set scope to current thread scope value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param scope pointer to thread scope value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getscope(const pthread_poolattr_t * attr,
        int * scope);

/**
 * @brief set current thread scope value to scope
 * @param attr pointer to pthread_pool_t attribute structure
 * @param scope thread scope value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setscope(pthread_poolattr_t * attr,
        const int scope);
#endif

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief set pshared to current process sharing value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getpshared(const pthread_poolattr_t * attr,
        int * pshared);

/**
 * @brief set current process sharing value to pshared
 * @param attr pointer to pthread_pool_t attribute structure
 * @param pshared process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setpshared(pthread_poolattr_t * attr,
        const int pshared);
#endif

#if (_POSIX_THREAD_ATTR_STACKSIZE == 1)
/**
 * @brief set stacksize to current thread stack size value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param stacksize pointer to thread stack size value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getstacksize(const pthread_poolattr_t * attr,
        size_t * stacksize);

/**
 * @brief set current thread stack size value to stacksize
 * @param attr pointer to pthread_pool_t attribute structure
 * @param stacksize thread stack size value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setstacksize(pthread_poolattr_t * attr,
        const size_t stacksize);
#endif

#ifdef CTF_DEBUG
#if ((_POSIX_THREAD_ATTR_STACKSIZE == 1) && \
        (_POSIX_THREAD_ATTR_STACKADDR == 1))
/**
 * @brief set checkstack to current check stack value
 * @param attr pointer to pthread_pool_t attribute structure
 * @param checkstack pointer to check stack value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_getcheckstack(const pthread_poolattr_t * attr,
        int * checkstack);

/**
 * @brief set current check stack value to checkstack
 * @param attr pointer to pthread_pool_t attribute structure
 * @param checkstack check stack value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_poolattr_setcheckstack(pthread_poolattr_t * attr,
        const int checkstack);
#endif
#endif

#ifdef __cplusplus
}
#endif

#endif
