/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief Propose Release Aquire Queue - praq (name by James Plourde)
 *
 * pthread_praq_t are functions that order output of results that
 * finish executing out of order
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_PRAQ_H_
#define _PTHREAD_PRAQ_H_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief enumeration for praq objects
 */
enum
{
    /**
     * @brief non-blocking enumeration
     */
    PTHREAD_PRAQ_NONBLOCKING,

    /**
     * @brief blocking enumeration.
     */
    PTHREAD_PRAQ_BLOCKING

};

/**
 * @brief non-blocking enumeration
 */
#define PTHREAD_PRAQ_NONBLOCKING PTHREAD_PRAQ_NONBLOCKING

/**
 * @brief blocking enumeration.
 */
#define PTHREAD_PRAQ_BLOCKING PTHREAD_PRAQ_BLOCKING

/**
 * @brief pthread_praq_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

    /**
     * @brief blocking attributes
     */
    int block;

} pthread_praqattr_t;


/**
 * @brief pthread_praq_t data structure
 */
struct praq_data_t
{
    /**
     * @brief data object
     */
    void * data;

    /**
     * @brief order id
     */
    int order_id;
};


/**
 * @brief pthread_praq_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief not full condition
     */
    pthread_cond_t not_full_cond;

    /**
     * @brief not empty condition
     */
    pthread_cond_t not_empty_cond;

    /**
     * @brief queue array
     */
    struct praq_data_t * queue;

    /**
     * @brief index of front of queue
     */
    int front;

    /**
     * @brief queue length
     */
    int length;

    /**
     * @brief maximum queue length
     */
    int max_length;

    /**
     * @brief count of threads blocking
     */
    int queue_full_wait;

    /**
     * @brief count of threads blocking
     */
    int queue_empty_wait;

    /**
     * @brief id of front of queue
     */
    int order_id;

    /**
     * @brief locking attribute
     */
    int block;

#ifdef CTF_STATS
    /**
     * @brief total count for queue
     */
    int total_count;

    /**
     * @brief total acquire wait count for queue
     */
    int total_acquire_wait_count;

    /**
     * @brief total propose wait count for queue
     */
    int total_propose_wait_count;

    /**
     * @brief sum of acquire time waited (sec)
     */
    float total_acquire_wait;

    /**
     * @brief sum of propose time waited (sec)
     */
    float total_propose_wait;
#endif

} pthread_praq_t;


/**
 * @brief initialize pthread_praq_t structure
 * @param praq pointer to pthread_praq_t structure
 * @param attr pointer to pthread_praq_t attribute structure
 * @param length maximum length of ordered queue
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_init(pthread_praq_t * praq,
        const pthread_praqattr_t * attr, const int length);

/**
 * @brief destroy pthread_praq_t structure
 * @param praq pointer to pthread_praq_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_destroy(pthread_praq_t * praq);

/**
 * @brief suspend calling thread until next data becomes available
 * @param praq pointer to pthread_praq_t structure
 * @param data pointer to data
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_acquire(pthread_praq_t * praq,
        void ** data);

/**
 * @brief suspend calling thread until order_num can be processed
 * @param praq pointer to pthread_praq_t structure
 * @param order_num order number
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_propose(pthread_praq_t * praq, const int order_num);

/**
 * @brief add data to queue for order order_num
 * @param praq pointer to pthread_praq_t structure
 * @param order_num order number
 * @param data pointer to data
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_release(pthread_praq_t * praq,
        const int order_num, void * data);

/**
 * @brief reset order back to zero
 * @param praq pointer to pthread_praq_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praq_reset(pthread_praq_t * praq);

/**
 * @brief initialize pthread_praq_t attribute structure
 * @param attr pointer to pthread_praq_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_init(pthread_praqattr_t * attr);

/**
 * @brief destroy pthread_praq_t attribute structure
 * @param attr pointer to pthread_praq_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_destroy(pthread_praqattr_t * attr);

/**
 * @brief assign block the current blocking attribute
 * @param attr pointer to pthread_praq_t attribute structure
 * @param block pointer to blocking attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_getblock(const pthread_praqattr_t * attr,
        int * block);

/**
 * @brief assign the current blocking attribute
 * @param attr pointer to pthread_praq_t attribute structure
 * @param block blocking attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_setblock(pthread_praqattr_t * attr,
        const int block);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief assign pshared the current process sharing attribute
 * @param attr pointer to pthread_praq_t attribute structure
 * @param pshared pointer to process sharing attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_getpshared(const pthread_praqattr_t * attr,
        int * pshared);

/**
 * @brief assign the current process sharing attribute
 * @param attr pointer to pthread_praq_t attribute structure
 * @param pshared process sharing attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_praqattr_setpshared(pthread_praqattr_t * attr,
        const int pshared);
#endif

#ifdef __cplusplus
}
#endif

#endif
