/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief functions that provide a read-write locking mechanism
 *
 * A barrier causes the calling thread to wait until a set number of
 * threads reach a point, then those threads may continue.
 * @warning <ul><li>not async-cancel safe</li></ul>
 * @warning <ul><li>cancellation has not been tested</li></ul>
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_RDRW_
#define _PTHREAD_RDRW_


#include <pthread.h>
#include <unistd.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief enumeration for rdwr preference
 */
enum
{
    /**
     * @brief reader preference enumeration
     */
    PTHREAD_RDWR_PREFER_READER,

    /**
     * @brief writer preference enumeration
     */
    PTHREAD_RDWR_PREFER_WRITER,

    /**
     * @brief toggle preference enumeration
     */
    PTHREAD_RDWR_PREFER_TOGGLE,

    /**
     * @brief default preference enumeration
     */
    PTHREAD_RDWR_PREFER_DEFAULT = PTHREAD_RDWR_PREFER_TOGGLE
};

/**
 * @brief reader preference enumeration
 */
#define PTHREAD_RDWR_PREFER_READER PTHREAD_RDWR_PREFER_READER

/**
 * @brief writer preference enumeration
 */
#define PTHREAD_RDWR_PREFER_WRITER PTHREAD_RDWR_PREFER_WRITER

/**
 * @brief toggle preference enumeration
 */
#define PTHREAD_RDWR_PREFER_TOGGLE PTHREAD_RDWR_PREFER_TOGGLE

/**
 * @brief default preference enumeration
 */
#define PTHREAD_RDWR_PREFER_DEFAULT PTHREAD_RDWR_PREFER_DEFAULT

/**
 * @brief enumeration for rdwr type
 */
enum
{
    /**
     * @brief normal type enumeration
     */
    PTHREAD_RDWR_NORMAL,

    /**
     * @brief error checking type enumeration
     */
    PTHREAD_RDWR_ERRORCHECK,

    /**
     * @brief default type enumeration
     */
    PTHREAD_RDWR_DEFAULT = PTHREAD_RDWR_NORMAL
};

/**
 * @brief normal type enumeration
 */
#define PTHREAD_RDWR_NORMAL PTHREAD_RDWR_NORMAL

/**
 * @brief error checking type enumeration
 */
#define PTHREAD_RDWR_ERRORCHECK PTHREAD_RDWR_ERRORCHECK

/**
 * @brief default type enumeration
 */
#define PTHREAD_RDWR_DEFAULT PTHREAD_RDWR_DEFAULT

/**
 * @brief pthread_rdwr_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

    /**
     * @brief process sharing attributes
     */
    int pshared;

    /**
     * @brief preference attribute
     */
    int preference;

    /**
     * @brief type attributes
     */
    int type;

} pthread_rdwrattr_t;


/**
 * @brief pthread_rdwr_t structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief read condition
     */
    pthread_cond_t read_cond;

    /**
     * @brief write condition
     */
    pthread_cond_t write_cond;

    /**
     * @brief number of readers
     */
    int reading;

    /**
     * @brief number of readers waiting
     */
    int readers_waiting;

    /**
     * @brief number of writers
     */
    int writing;

    /**
     * @brief number of writers waiting
     */
    int writers_waiting;

    /**
     * @brief current stage of rdwr-lock (only for toggle preference)
     *
     * -2 -> stage writing
     * -1 -> stage waiting for writing
     *  0 -> stage open
     *  1 -> stage waiting for reading
     *  2 -> stage reading
     */
    short stage;

    /**
     * @brief preference attribute
     */
    int preference;

    /**
     * @brief type attribute
     */
    int type;

#ifdef CTF_STATS
    /**
     * @brief total count of readers
     */
    int total_read_pass_count;

    /**
     * @brief total count of readers that waited
     */
    int total_read_wait_count;

    /**
     * @brief total count of writers
     */
    int total_write_pass_count;

    /**
     * @brief total count of writers that waited
     */
    int total_write_wait_count;

    /**
     * @brief sum of time reading (sec)
     */
    float total_read_wait;

    /**
     * @brief sum of time writing (sec)
     */
    float total_write;

    /**
     * @brief sum of time waiting for writing (sec)
     */
    float total_write_wait;
#endif

} pthread_rdwr_t;


/**
 * @brief static initializer for pthread_rdwr_t
 */
#ifdef CTF_STATS
#define PTHREAD_RDWR_INITIALIZER { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER, \
        0, 0, 0, 0, 0, PTHREAD_RDWR_PREFER_DEFAULT, PTHREAD_RDWR_DEFAULT, \
        0, 0, 0, 0, 0.0, 0.0, 0.0 }
#else
#define PTHREAD_RDWR_INITIALIZER { PTHREAD_MUTEX_INITIALIZER, \
        PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER, \
        0, 0, 0, 0, 0, PTHREAD_RDWR_PREFER_DEFAULT, PTHREAD_RDWR_DEFAULT }
#endif


/**
 * @brief initialize pthread_rdwr_t structure
 * @param rdwr pointer to pthread_rdwr_t structure
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_init(pthread_rdwr_t * rdwr,
        const pthread_rdwrattr_t * attr);

/**
 * @brief initialize pthread_rdwr_t structure
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_destroy(pthread_rdwr_t * rdwr);

/**
 * @brief to obtain a read lock
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_rdlock(pthread_rdwr_t * rdwr);

/**
 * @brief to try to obtain a read lock for at least timeout duration
 * @param rdwr pointer to pthread_rdwr_t structure
 * @param timeout pointer to timespec structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_timedrdlock(pthread_rdwr_t * rdwr,
        const struct timespec * timeout);

/**
 * @brief to obtain a read lock if no write lock
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_tryrdlock(pthread_rdwr_t * rdwr);

/**
 * @brief to obtain a write lock
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_wrlock(pthread_rdwr_t * rdwr);

/**
 * @brief try to obtain a write lock for at least timeout duration
 * @param rdwr pointer to pthread_rdwr_t structure
 * @param timeout pointer to timespec structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_timedwrlock(pthread_rdwr_t * rdwr,
        const struct timespec * timeout);

/**
 * @brief try to obtain a write lock if no read lock or write lock
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_trywrlock(pthread_rdwr_t * rdwr);

/**
 * @brief unlock current thread
 * @param rdwr pointer to pthread_rdwr_t structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwr_unlock(pthread_rdwr_t * rdwr);

/**
 * @brief initialize pthread_rdwr_t attribute structure
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_init(pthread_rdwrattr_t * attr);

/**
 * @brief destroy pthread_rdwr_t attribute structure
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_destroy(pthread_rdwrattr_t * attr);

/**
 * @brief assign kind the current kind attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param pref pointer to kind attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_getkind(const pthread_rdwrattr_t * attr,
        int * pref);

/**
 * @brief assign current kind attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param pref kind attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_setkind(pthread_rdwrattr_t * attr, const int pref);

/**
 * @brief assign type the current type attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param type pointer to type attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_gettype(const pthread_rdwrattr_t * attr,
        int * type);

/**
 * @brief assign current type attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param type type attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_settype(pthread_rdwrattr_t * attr, const int type);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief assign pshared the current process sharing attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param pshared pointer to process sharing attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_getpshared(const pthread_rdwrattr_t * attr,
        int * pshared);

/**
 * @brief assign current process sharing attribute
 * @param attr pointer to pthread_rdwr_t attribute structure
 * @param pshared process sharing attribute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_rdwrattr_setpshared(pthread_rdwrattr_t * attr,
        const int pshared);
#endif

#ifdef __cplusplus
}
#endif

#endif
