/* 
 * Copyright (C) 2005 Abel Bennett.
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 */

/**
 * @file
 * @brief functions that provide a sub pool of pthread_pool_t
 */

/**
 * @author Abel Bennett
 * @date 2003/05/08
 * @version 1.0
 */


#ifndef _PTHREAD_SUBPOOL_
#define _PTHREAD_SUBPOOL_


#include <pthread_pool.h>


#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief pthread_subpool_t attribute structure
 */
typedef struct
{
    /**
     * @brief mutex attributes
     */
    pthread_mutexattr_t mutex_attr;

    /**
     * @brief condition attributes
     */
    pthread_condattr_t cond_attr;

} pthread_subpoolattr_t;


/**
 * @brief subpool structure
 */
typedef struct
{
    /**
     * @brief structure lock
     */
    pthread_mutex_t lock;

    /**
     * @brief work finished condition
     */
    pthread_cond_t cond;

    /**
     * @brief pointer to thread pool
     */
    pthread_pool_t * pool;

    /**
     * @brief number of jobs
     */
    int jobs;

} pthread_subpool_t;


/**
 * @brief initialize subpool structure
 * @param subpool pointer to subpool structure
 * @param attr pointer to subpool attribute structure
 * @param pool pointer to pool strucute
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpool_init(pthread_subpool_t * subpool,
        const pthread_subpoolattr_t * attr,
        pthread_pool_t * pool);

/**
 * @brief destroy subpool structure
 * @param subpool pointer to subpool structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpool_destroy(pthread_subpool_t * subpool);

/**
 * @brief add routine(arg) to subpool work to be done
 * @param subpool pointer to subpool structure
 * @param routine pointer to a function
 * @param arg pointer to argument for routine
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpool_addwork(pthread_subpool_t * subpool,
        void (*routine)(void *), void * arg);

/**
 * @brief signal the subpool that a job has finished
 * @param subpool pointer to subpool structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpool_finishjob(pthread_subpool_t * subpool);

/**
 * @brief suspend calling thread until all subpool work is done
 * @param subpool pointer to subpool structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpool_finishwork(pthread_subpool_t * subpool);

/**
 * @brief initialize subpool attribute structure
 * @param attr pointer to subpool attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpoolattr_init(pthread_subpoolattr_t * attr);

/**
 * @brief destroy subpool attribute structure
 * @param attr pointer to subpool attribute structure
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpoolattr_destroy(pthread_subpoolattr_t * attr);

#if (_POSIX_THREAD_PROCESS_SHARED == 1)
/**
 * @brief set pshared to subpool process sharing value
 * @param attr pointer to subpool attribute structure
 * @param pshared pointer to process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpoolattr_getpshared(const pthread_subpoolattr_t * attr,
        int * pshared);

/**
 * @brief set subpool process sharing value to pshared
 * @param attr pointer to subpool attribute structure
 * @param pshared process sharing value
 * @return
 * 0 is returned on success<br>
 * non-zero error code is returned on failure
 */
extern int pthread_subpoolattr_setpshared(pthread_subpoolattr_t * attr,
        const int pshared);
#endif

#ifdef __cplusplus
}
#endif

#endif
